"""
This module contains the class ToolFunction.

Author:
    Yu-Ren Liu, Yang Yu
"""

import pickle


class ToolFunction:
    """
    Class ToolFunction defines some tool function used in project.
    """
    def __init__(self):
        pass

    @staticmethod
    def list_compare(list1, list2):
        """
        Compare two lists. If lists are same, return True.

        :return: True or False
        """
        if len(list1) != len(list2):
            return False
        for i in range(len(list1)):
            if list1[i] != list2[i]:
                return False
        return True

    @staticmethod
    def deepcopy(obj):
        """
        Deep copy an object.

        :param obj: object
        :return: a new object
        """
        return pickle.loads(pickle.dumps(obj));

    @staticmethod
    def log(text):
        """
        Log output.

        :param text: text to output
        :return: no return
        """
        print('[zoojl] '+text)



"""
This module contains the class Global.

Author:
    Yu-Ren Liu
"""

from random import Random


class Global:
    """
    This class defines global variables used in the algorithm.
    """
    rand = None

    def __init__(self):
        """
        Initialization, includes rand and precision.
        """
        # rand is the random object used by all files
        self.rand = Random()
        self.precision = 1e-17
        # rand.seed(100)

    def set_seed(self, seed):
        """
        Set random seed.

        :param seed: random seed
        :return: no return
        """
        self.rand.seed(seed)
        return

    def set_precision(self, my_precision):
        """
        Set precision, precision is used to judge whether two floats are equal.

        :param my_precision: precision
        :return: no return
        """
        self.precision = my_precision
        return

gl = Global()
# constants
pos_inf = float('Inf')
neg_inf = float('-Inf')
nan = float('Nan')



"""
You can run this file to start the evaluation servers.

Author:
    Yu-Ren Liu
"""

import os
import sys
import socket
from zoosrv.components.solution import Solution
from zoosrv.components.loader import Loader
from zoosrv.components.receive import receive
from zoosrv.components.tool_function import ToolFunction
from zoosrv.components.port_conflict import is_open
import multiprocessing
import ConfigParser


class EvaluationServer:
    """
        Evaluation server ,an important part in asynchronous racos, is responsible for
        evaluating solution sent by client.
    """
    def __init__(self, s_ip, s_port, data_len):
        """
        Initialization.

        :param s_ip: server ip
        :param s_port: server port
        :param data_len: data length in tcp
        """
        self.__server_ip = s_ip
        self.__server_port = s_port
        self.__data_length = data_len

        return

    def start_server(self, control_server, shared_fold):
        """
        Start this evaluation server.

        :param control_server: control server address
        :param shared_fold: current working directory
        :return: target function name
        """

        # send evaluation server address to control server
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.explain_address(control_server)))
        s.sendall("evaluation server#")
        receive(self.__data_length, s)
        s.sendall(self.__server_ip + ':' + str(self.__server_port) + "#")
        s.close()

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((self.__server_ip, self.__server_port))
        s.listen(5)
        all_connect = 0
        restart = False
        ToolFunction.log("Initialize evaluation process successfully: ip=%s, port=%s" % (self.__server_ip, self.__server_port))
        while True:
            # get x from client
            es, address = s.accept()
            # print all_connect + 1, ' get connected...'
            all_connect += 1
            cmd = receive(self.__data_length, es)
            if cmd == "control server: shutdown":
                es.sendall("success#")
                break
            elif cmd == "client: calculate":
                es.sendall("calculate\n")
                try:
                    msg = receive(self.__data_length, es)
                    es.sendall("receive\n")
                    if msg == "pposs":
                        msg = receive(self.__data_length, es)
                        addr, func, constraint = msg.split(":")
                        es.sendall("receive\n")
                        load = Loader()
                        module = load.load(shared_fold + addr)
                        calculate_fx = module[func]
                        calculate_constraint = module[constraint]
                        data = receive(self.__data_length, es)
                        x = []
                        data_str = data.split(' ')
                        for istr in data_str:
                            x.append(float(istr))
                        fx = calculate_fx(Solution(x=x))
                        c = calculate_constraint(Solution(x=x))
                        fx_x = str(fx) + ' ' + str(c) + "\n"
                        es.sendall(fx_x)
                        msg = "Server:%s:%s f(x):%s constraint:%s connect num:%s client address:%s" % (
                            self.__server_ip, self.__server_port, fx, c, all_connect, address)
                        ToolFunction.log(msg)
                    elif msg == "asracos":
                        msg = receive(self.__data_length, es)
                        addr, func = msg.split(":")
                        es.sendall("receive\n")
                        load = Loader()
                        module = load.load(shared_fold + addr)
                        calculate = module[func]
                        data = receive(self.__data_length, es)
                        x = []
                        data_str = data.split(' ')
                        for istr in data_str:
                            x.append(float(istr))
                        fx = calculate(Solution(x=x))
                        fx_x = str(fx) + "\n"
                        es.sendall(fx_x)
                        msg = "Server:%s:%s f(x):%s connect num:%s client address:%s" % (
                            self.__server_ip, self.__server_port, fx, all_connect, address)
                        ToolFunction.log(msg)
                    else:
                        ToolFunction.log("Exception: %s method is not implemented" % msg)
                except Exception, msg:
                    ToolFunction.log("Exception")
                    es.sendall("Exception: " + str(msg))
                    restart = True
                    break
                # print ("send result finished, result: " + str(fx_x))
            elif cmd == "control server: restart":
                restart = True
                break
            else:
                ToolFunction.log("Command error: no such cmd")
            es.close()
        ToolFunction.log("Server close!")
        s.close()
        if restart is True:
            ToolFunction.log("Server restart")
            self.start_server(control_server, shared_fold)

    def explain_address(self, addr):
        """
        Get ip and port from ad`dress 'ip:port'. Ip is a string and port is a integer.

        :param addr: address
        :return: ip and port
        """
        addr = addr.split(':')
        t_ip = addr[0]
        t_port = int(addr[1])
        return t_ip, t_port


def run(port, shared_fold, control_server):
    """
    Api of running evaluation server.

    :param port: port of evaluation server
    :param shared_fold: the working directory shared by julia client and evaluation servers
    :param control_server: ip:port of the control server
    :return: no return value
    """
    local_ip = socket.gethostbyname(socket.gethostname())
    data_length = 1024
    server_ip = local_ip
    server_port = port

    # set server ip, port and longest data length in initialization
    server = EvaluationServer(server_ip, server_port, data_length)

    server.start_server(control_server=control_server, shared_fold=shared_fold)


def start(configuration):
    """
    Starting evaluation servers from configuration file.

    :param configuration:
        configuration is a file name
        configuration  has five lines
        he first line is the directory this server works under, objective functions should be defined in this directory
        the second line is the address of control server
        the third line is the number of evaluation servers to be started
        the fourth line is the lower bound of the range that ports are chosen from
        the last line is the upper bound of the range that ports are chosen from
        2 means opening 2 server, 60003 and 60010 mean these servers can use port between 60003 and 60010([60003, 60010])
    :return: no return value
    """
    conf = ConfigParser.ConfigParser()
    conf.read(configuration)
    section = conf.sections()[0]
    options = conf.options(section)
    shared_fold = conf.get(section, "shared fold")
    control_server = conf.get(section, "control server ip_port")
    num = conf.getint(section, "evaluation processes")
    starting_port = conf.getint(section, "starting port")
    ending_port = conf.getint(section, "ending port")
    local_ip = socket.gethostbyname(socket.gethostname())  # get local ip
    count = 0
    workers = []
    for port in range(starting_port, ending_port):
        if is_open(local_ip, port) is False:
            count += 1
            workers.append(multiprocessing.Process(target=run, args=(port, shared_fold, control_server)))
            if count >= num:
                break
    for w in workers:
        w.start()




"""
This file contains a function judging whether a port is occupied.

Author:
    Yu-Ren Liu
"""

import socket


def is_open(ip, port):
    """
    Judge whether a port is occupied or not.

    :param ip: ip address
    :param port: port number
    :return: True or False
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((ip, port))
    if result == 0:
        print("Port %d is already in use")
        return True
    else:
        return False




"""
This module contains the class Loader, which can load another module from its path.

Author:
    Yu-Ren Liu
"""

import traceback
import sys


class Loader(object):
    """
    This class can load a module from path.
    """

    def __init__(self):
        pass

    def load(self, path):
        """
        Load a module from path.

        :param path: path of the module
        :return: loaded module(succeed) or None(fail)
        """
        try:
            tmp = {}
            exec open(path).read() in tmp
            return tmp
        except:
            print("Load module [path %s] error: %s"
                  % (path, traceback.format_exc()))
            return None


"""
You can run this file to start the control server.

Author:
    Yu-Ren Liu
"""

import threading
import socket
from zoosrv.components.receive import receive
from zoosrv.components.tool_function import ToolFunction
import copy


class ControlServer:
    """
    Control server is responsible for managing all evaluation servers.
    """
    def __init__(self, ip, port):
        """
        Initialization.

        :param ip: ip address of the control server, e.g., "127.0.0.1"
        :param port:
            ports occupied by the control server, it's a list having four elements, e.g., [10000, 10001, 10002, 10003]
            respectively are occupied by function receive_from_evaluation_server, send_to_client and receive_from_client, restart_evaluation_server
        """
        self.ip = ip
        self.port = port
        # for example, elements are ["192.168.1.101:5555", "192.168.1.102:10000"]
        self.evaluation_server = []

    def start(self):
        """
        Start this control server.
        :return: no return
        """
        lock = threading.RLock()
        # start threads
        worker = []
        worker.append(threading.Thread(target=self.communication))
        for t in worker:
            t.setDaemon(True)
            t.start()
        # start main thread
        while True:
            cmd = int(raw_input("[zoojl] Please input command sequence number, 1: print evaluation servers, 2: "
                                "shut down evaluation servers, 3: exit\n"))
            with lock:
                if cmd == 1:
                    ToolFunction.log("The number of evaluation servers: %s" % len(self.evaluation_server))
                    ToolFunction.log(str(self.evaluation_server))
                elif cmd == 2:
                    self.shut_down_control()
                elif cmd == 3:
                    return

    def communication(self):
        """
        main process
        :return:
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)  #
        s.bind((self.ip, self.port))
        s.listen(20)
        while True:
            client, address = s.accept()
            cmd = receive(1024, client)
            client.sendall("success#\n")
            # receive from evaluation servers
            if cmd == "evaluation server":
                ip_port = receive(1024, client)
                if ip_port not in self.evaluation_server:
                    self.evaluation_server.append(ip_port)
                    ToolFunction.log("Receive ip_port from evaluation server: " + ip_port)
            # send to client
            elif cmd == "client: require servers":
                require_num = int(receive(1024, client))
                require_num = min(require_num, len(self.evaluation_server))
                msg = ""
                for i in range(require_num):
                    if i == 0:
                        msg += self.evaluation_server[i]
                    else:
                        msg = msg + ' ' + self.evaluation_server[i]
                if require_num == 0:
                    msg += "No available evaluation process"
                msg += '\n'
                self.evaluation_server = self.evaluation_server[require_num:]
                if require_num <= 1:
                    client.sendall('Get %d evaluation process\n' % require_num)
                else:
                    client.sendall('Get %d evaluation processes\n' % require_num)
                client.sendall(msg)
            # receive from client
            elif cmd == "client: return servers":
                data_str = receive(1024, client)
                ToolFunction.log("Receive ip_port from client: " + data_str)
                ip_ports = data_str.split()
                for ip_port in ip_ports:
                    if ip_port not in self.evaluation_server:
                        self.evaluation_server.append(ip_port)
            # exception happens at client. restart evaluation servers
            elif cmd == "client: restart":
                data_str = receive(1024, client)
                ToolFunction.log("Client exception, receive ip_port from client: " + data_str)
                ip_ports = data_str.split()
                for ip_port in ip_ports:
                    if ip_port not in self.evaluation_server:
                        ip, port = ip_port.split(":")
                        port = int(port)
                        addr = (ip, port)
                        s.connect(addr)
                        s.sendall("control server: restart\n")
                        self.evaluation_server.append(ip_port)
            else:
                ToolFunction.log("Command error")

    def shut_down_control(self):
        """
        Shut down evaluation servers.

        :return: no return
        """
        try:
            ToolFunction.log("you can input three different kinds of commands")
            ToolFunction.log("1.all ==> shut down all servers. e.g. all")
            ToolFunction.log("2.ip:ip1 ==> shut down all servers having ip1 as ip. e.g. ip:127.0.0.1")
            ToolFunction.log("3.ip_port: ip1:port1 ip2:port2 ip3:port3 ... ==> shut down specific servers. e.g. ip_port: 127.0.0.1:20000 127.0.0.1:20001")
            msg = raw_input("")
            new_cal_server = copy.deepcopy(self.evaluation_server)
            if msg == "all":
                for ip_port in self.evaluation_server:
                    self.shut_down(ip_port)
                    new_cal_server.remove(ip_port)
            elif msg[:7] == "ip_port":
                ip_ports = msg[9:].split(' ')
                for ip_port in ip_ports:
                    self.shut_down(ip_port)
                    new_cal_server.remove(ip_port)
            elif msg[:2] == "ip":
                ip = msg[3:]
                for ip_port in self.evaluation_server:
                    sip, port = ip_port.split(":")
                    if sip == ip:
                        self.shut_down(ip_port)
                        new_cal_server.remove(ip_port)
            else:
                ToolFunction.log("No such command")
            self.evaluation_server = new_cal_server
        except Exception as e:
            ToolFunction.log("input error", e)

    def shut_down(self, ip_port):
        """
        Shut down one evaluation server, which bounds to ip_port.

        :param ip_port: ip:port of the evaluation server
        :return: no return
        """
        es = socket.socket(socket.AF_INET, socket.SOCK_STREAM)  #
        if ip_port not in self.evaluation_server:
            ToolFunction.log("no such ip:port")
            return
        ip, port = ip_port.split(":")
        port = int(port)
        addr = (ip, port)
        es.connect(addr)
        es.sendall("control server: shutdown#")
        result = receive(1024, es)
        if result == "success":
            ToolFunction.log("%s: manage to shut down" % ip_port)
        else:
            ToolFunction.log("fail to shut down")
        es.close()


def start(port):
    """
    Start the control server.

    :param port:
        The port occupied by the control server
    :return: no return
    """
    local_ip = socket.gethostbyname(socket.gethostname())
    ToolFunction.log("Control server ip_port: %s:%s" %(local_ip, port))
    cs = ControlServer(local_ip, port)
    cs.start()



"""
This module contains the class Solution.

Author:
    Yu-Ren Liu
"""

from zoosrv.components.zoo_global import pos_inf, neg_inf, nan, gl
from zoosrv.components.tool_function import ToolFunction


class Solution:
    """
    A solution encapsulates a solution vector with attached properties, including dimension information, objective value,
and attachment
    """
    def __init__(self, x=[], value=nan, resample_value=None, attach=None, post_attach=None, is_in_possible_solution=False):
        """
        Initialization.

        :param x: a list
        :param value: objective value
        :param resample_value: reevaluated value
        :param attach: attached structure
        :param post_attach:
        :param is_in_possible_solution:
        """
        self.__x = x
        self.__value = value
        self.__resample_value = resample_value
        self.__attach = attach
        self.__post_attach = post_attach
        self.__is_in_possible_solution = is_in_possible_solution
        return

    @property
    def is_in_possible_solution(self):
        return self.__is_in_possible_solution

    @is_in_possible_solution.setter
    def is_in_possible_solution(self, value):
        self.__is_in_possible_solution = value

    def deep_copy(self):
        """
        Deep copy this solution. Note that the attachment is not deeply copied
        :return: a new solution
        """

        x = []
        for x_i in self.__x:
            x.append(x_i)
        value = self.__value
        attach = self.__attach
        resample_value = self.__resample_value
        post_attach = self.__post_attach
        return Solution(x, value, resample_value, attach, post_attach, self.is_in_possible_solution)

    def is_equal(self, sol):
        """
        Check if two solutions equal

        :param sol: another solution
        :return: True or False
        """
        sol_x = sol.get_x()
        sol_value = sol.get_value()
        if sol_value != nan and self.__value != nan:
            if abs(self.__value - sol_value) > gl.precision:
                return False
        if len(self.__x) != len(sol_x):
            return False
        for i in range(len(self.__x)):
            if abs(self.__x[i] - sol_x[i]) > gl.precision:
                return False
        return True

    def exist_equal(self, sol_set):
        """
        Check if exists another solution in sol_set the same as this one

        :param sol_set: solution set
        :return: True or False
        """
        for sol in sol_set:
            if self.is_equal(self, sol):
                return True
        return False

    def set_x_index(self, index, x):
        self.__x[index] = x
        return

    def set_x(self, x):
        self.__x = x
        return

    def set_value(self, value):
        self.__value = value
        return

    def set_attach(self, attach):
        self.__attach = attach
        return

    def set_post_attach(self, attach):
        self.__post_attach = attach
        return

    def set_resample_value(self, resample_value):
        self.__resample_value = resample_value

    def get_resample_value(self):
        return self.__resample_value

    def get_post_attach(self):
        return self.__post_attach

    def get_x_index(self, index):
        return self.__x[index]

    def get_x(self):
        return self.__x

    def get_value(self):
        return self.__value

    def get_attach(self):
        return self.__attach

    def print_solution(self):
        ToolFunction.log('x: ' + repr(self.__x))
        ToolFunction.log('value: ' + repr(self.__value))

    @staticmethod
    def deep_copy_set(sol_set):
        """
        Deep copy an solution set

        :param sol_set: solution set
        :return: copied solution set
        """
        result_set = []
        for sol in sol_set:
            result_set.append(sol.deep_copy())
        return result_set

    @staticmethod
    def print_solution_set(sol_set):
        """
        Print the value of each solution in an solution set.

        :param sol_set: solution set
        :return: no return
        """
        for sol in sol_set:
            ToolFunction.log('value: %f' % (sol.get_value()))
        return

    @staticmethod
    def find_maximum(sol_set):
        """
        Find the solution having maximum value from the solution set

        :param sol_set: solution set
        :return: solution, index
        """
        maxi = neg_inf
        max_index = 0
        for i in range(len(sol_set)):
            if sol_set[i].get_value() > maxi:
                maxi = sol_set[i].get_value()
                max_index = i
        return sol_set[max_index], max_index

    @staticmethod
    def find_minimum(sol_set):
        """
        Find the solution having minimum value from the solution set

        :param sol_set: solution set
        :return: solution, index
        """
        mini = pos_inf
        mini_index = 0
        for i in range(len(sol_set)):
            if sol_set[i].get_value() < mini:
                mini = sol_set[i].get_value()
                mini_index = i
        return mini, mini_index



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='zoosrv',
    version="0.1.1",
    description=(
        'Python Servers for Distributed Zeroth-Order Optimization'
    ),
    author='Yang Yu',
    author_email='yuy@nju.edu.cn',
    maintainer='Yu-Ren Liu, Yang Yu',
    license='MIT License',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/eyounx/ZOOsrv',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'ConfigParser',
        'multiprocessing',
    ]
)



"""
This module contains the function 'receive', which receives data from evaluation server.

Author:
    Yu-Ren Liu
"""


def receive(data_length, es):
    """
    Receive data from evaluation server and remove useless characters.

    :param data_length: length of the data received
    :param es: evaluation server
    :return: data received
    """
    data_str = ""
    while True:
        data_tmp = es.recv(data_length)
        stop = data_tmp.find("#")
        if stop > 0:
            data_tmp = data_tmp[0:stop]
            data_str += data_tmp
            break
        else:
            data_str += data_tmp
    if data_str[0] == '\n':
        data_str = data_str[1:]
    if data_str[-1] == '\n':
        data_str = data_str[:-1]
    return data_str
